/*
 *  Fireworks OpenGL Demonstrator
 *  Moving parallel or spot light with textures and ambient light control
 *
 *  Created by brisson on 2012/05/17.
 *  Updated by brisson on 2013/05/04.
 *  Copyright (c) 2012-2013 J M. Brisson Lopes & Instituto Superior Tecnico. All rights reserved.
 *
 */
/******************************************************************************
* Fireworks - an OpenGL program to simulate fireworks explosions              *
* Copyright (C) 2012, 2013 IST & J M Brisson Lopes                            *
*                                                                             *
*  This software is provided 'as-is', without any express or implied          *
*  warranty.  In no event will the authors be held liable for any damages     *
*  arising from the use of this software.                                     *
*                                                                             *
*  Permission is granted to anyone to use this software for any purpose,      *
*  including commercial applications, and to alter it and redistribute it     *
*  freely, subject to the following restrictions:                             *
*                                                                             *
*  1. The origin of this software must not be misrepresented; you must not    *
*     claim that you wrote the original software. If you use this software    *
*     in a product, an acknowledgment in the product documentation would be   *
*     appreciated but is not required.                                        *
*  2. Altered source versions must be plainly marked as such, and must not be *
*     misrepresented as being the original software.                          *
*  3. This notice may not be removed or altered from any source distribution. *
*                                                                             *
* The "official" glbmp webpage is: http://chaoslizard.sourceforge.net/glbmp/  *
******************************************************************************/


#if defined(__APPLE__) || defined(MACOSX)
#include <GLUT/glut.h>
#else
#include <GL/glut.h>
#endif

#include "Fireworks.h"
#include "glbmp.h"
#include <iostream>
#include <sstream>
#include <iomanip>
#include <cstdlib>
#include <ctype.h>
#include <cmath>
#include <cstdlib>

using namespace std;

#define frand()			((float)rand()/RAND_MAX)
#ifndef M_PI
#define M_PI 3.14159265
#endif



// Particle implementation

Particle::Particle()
{
    life = 0.0f;
    Respawn();
}


const bool Particle::Respawn( const bool force)
{
    if ( life <= 0.0f  || force )
    {
        double v = 0.8*frand()+0.2;
        double phi = frand()*M_PI;
        double theta = 2.0*frand()*M_PI;
        x = 0.0f;
        y = 15.0f;
        z = 0.0f;
        vx = v * cos(theta) * sin(phi);
        vy = v * cos(phi);
        vz = v * sin(theta) * sin(phi);
        ax =  0.01f; /* just a little wind */
        ay = -0.15f; /* gravity pull */
        az =  0.0f;
        
        /* yelowish colour to be multiplied by the texture grey tone */
        r = 0.882f;
        g = 0.552f;
        b = 0.211f;
        
        life = 1.0f;		                        /* life granted to particle */
		fade = 0.0001515 + frand() * .000303;	    /* life decay */
        return true;
    }
    return false;
}


const Particle & Particle::operator = (const Particle & another)
{
    x  = another.x;
    y  = another.y;
    z  = another.z;
    vx = another.vx;
    vy = another.vy;
    vz = another.vz;
    ax = another.ax;
    ay = another.ay;
    az = another.az;
    r  = another.r;
    g  = another.g;
    b  = another.b;
    life = another.life;
    fade = another.fade;
    
    return *this;
}


const bool Particle::Update(const float elapsedTime, const bool respawn)
{
    if ( respawn && life <= 0. )
        Respawn();
    if ( life > 0. )
    {
        float delta = elapsedTime * .00378;
        x += (vx * delta);
        y += (vy * delta);
        z += (vz * delta);
        vx += (ax * delta);
        vy += (ay * delta);
        vz += (az * delta);
        life -= (fade * elapsedTime );
        return true;
    }
    return false;
}


unsigned int Particle::ParticleSkeleton;


void Particle::Draw() const
{
    glPushMatrix();
    glTranslatef( x, y, z);
    glColor4f( r, g, b, life);
    glCallList( ParticleSkeleton);
    glPopMatrix();
}


unsigned int Particle::CreateParticleSkeleton()
{
    ParticleSkeleton = glGenLists( 1);
    glNewList( ParticleSkeleton, GL_COMPILE);
    glScalef( .5, .5, 1.);
    glBegin(GL_TRIANGLE_STRIP);	// 4 vertices - 2 triangles: v0 v1 v2; v1 v2 v3 */
    glTexCoord2d(1,1);
    glVertex3f( +1.0f, +1.0f, 0.f);
    glTexCoord2d(0,1);
    glVertex3f( -1.0f, +1.0f, 0.f);
    glTexCoord2d(1,0);
    glVertex3f( +1.0f, -1.0f, 0.f);
    glTexCoord2d(0,0);
    glVertex3f( -1.0f, -1.0f, 0.f);
    glEnd();
    glEndList();
    return ParticleSkeleton;
}


unsigned int Particle::LoadTexture( const string &textureBitmapFilename)
{
    glbmp_t bitmap;     // use GLMBMP to load texture
    
    //try to load the specified file--if it fails, dip out
    if(!glbmp_LoadBitmap( textureBitmapFilename.c_str(), 0, &bitmap))
    {
        cerr << "Error loading bitmap file: " << textureBitmapFilename << endl;
        exit(1);
    }
    
    unsigned int texname;
	glGenTextures(1, &texname);
	glBindTexture(GL_TEXTURE_2D, texname);
	//copy data from bitmap into texture
	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, bitmap.width, bitmap.height,
				 0, GL_RGB, GL_UNSIGNED_BYTE, bitmap.rgb_data);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
    glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE); /* unnecessary, this is the default */
	//free the bitmap
	glbmp_FreeBitmap(&bitmap);
    return texname;
}


// ParticleSystem Implementation


ParticleSystem::ParticleSystem( const bool burst)
{
    for ( int i=0; i<Max_Particles; i++)
        activeParticles.push_back( new Particle());
    respawnParticles = !burst;
}


int ParticleSystem::Max_Particles = 1000;


void ParticleSystem::Update( const float elapsed)
{
    list<Particle*>::iterator it = activeParticles.begin();
    while ( it != activeParticles.end() ) {
        (*it)->Update( elapsed, respawnParticles);
		if ( (*it)->IsAlive() )
			it++;
		else
		{
			deadParticles.push_back( *it);
			it = activeParticles.erase( it);
		}
    }
}


void ParticleSystem::Draw() const
{
    list<Particle*>::const_iterator it = activeParticles.begin();
    while ( it != activeParticles.end() ) {
        (*it)->Draw();
        it++;
    }
}


void ParticleSystem::Reset()
{
    list<Particle*>::iterator it = activeParticles.begin();
    while ( it != activeParticles.end() ) {
        (*it)->Respawn( true);
        it++;
    }
	it = deadParticles.begin();
	while( it != deadParticles.end() )
	{
		(*it)->Respawn();
		activeParticles.push_back( *it);
		it = deadParticles.erase( it);
	}
}


// Camera implementation

Camera::Camera()
{
    Projection();
    glMatrixMode( GL_MODELVIEW);
    glLoadIdentity();
    gluLookAt (0.0, 0.0, 80., 0.0, 0.0, 0.0, 0.0, 1.0, 0.0);
    
}


float Camera::CameraFOV = 45.;


void Camera::Projection() const
{
    glMatrixMode( GL_PROJECTION);
    glLoadIdentity();
    gluPerspective( CameraFOV, (float) glutGet( GLUT_WINDOW_WIDTH) / glutGet( GLUT_WINDOW_HEIGHT),
                    1., 100.);
}


// OpenGLApplication implementation *******************************************

OpenGLApplication::OpenGLApplication()
{
    TimerOn    = true;
    ShowHUD    = true;
    Fullscreen = false;
}


OpenGLApplication OpenGLApplication::MyFireworks;


void OpenGLApplication::InitApplication( int argc, char **argv)
{
    glutInit( &argc, argv);
    glutInitDisplayMode( GLUT_RGB | GLUT_SINGLE | GLUT_DEPTH);
    glutInitWindowSize( 640, 480);
    glutCreateWindow( "Fireworks (C)2013 IST & JMBL");
    
    glClearColor( 0., 0., 0., 0.);
    
    MyCamera = new Camera();
    MySystem = new ParticleSystem( true);
    
    glutReshapeFunc( reshape);
    glutDisplayFunc( display);
    glutKeyboardFunc( keyboard);
    glutTimerFunc( 33, timer, 1);
    
    glEnable( GL_TEXTURE_2D);
    glEnable(GL_BLEND);
    glBlendFunc(GL_SRC_ALPHA,GL_ONE);
    glHint(GL_PERSPECTIVE_CORRECTION_HINT,GL_NICEST);
    Particle::CreateParticleSkeleton();
    Particle::LoadTexture( "particula.bmp");
    
    lastTime                = glutGet( GLUT_ELAPSED_TIME);
    lastFramerateUpdateTime = lastTime;
    frameConter             = 0;
    
    glutMainLoop();
}


void OpenGLApplication::MyReshape(const int width, const int height)
{
    glViewport( 0, 0, width, height);
    MyCamera->Projection();
}


void OpenGLApplication::MyDisplay()
{
    // How long has it been since last update ?
    int now    = glutGet( GLUT_ELAPSED_TIME);
    int elaped = now - lastTime;
    lastTime   = now;
    
    // Compute frame rate
    frameConter++;
    if ( now >= (lastFramerateUpdateTime + 1000) )
    {
        frameRate               = (1000. * frameConter) / (now - lastFramerateUpdateTime);
        lastFramerateUpdateTime = now;
        frameConter             = 0;
    }
    
    // Update particles
    MySystem->Update( elaped);
 
    glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
   
    MySystem->Draw();
    
    if ( ShowHUD)
        HUD();
    
    glFlush();
}


void OpenGLApplication::HUD() const
{
    // Switch to raster mode (saving active projection)
    glDisable( GL_TEXTURE_2D);
    glMatrixMode( GL_PROJECTION);
    glPushMatrix();
    glLoadIdentity();
    gluOrtho2D( 0, glutGet( GLUT_WINDOW_WIDTH), 0, glutGet( GLUT_WINDOW_HEIGHT));
    glMatrixMode( GL_MODELVIEW);
    glPushMatrix();
    glLoadIdentity();
    
    glColor3f( 0., 1., 0.);
    DisplayString( 5, 5, CommandLabel);
    
    ostringstream os;
    os << setprecision( 1) << "Frame rate: " << fixed << setw(6) << frameRate
       << " fps; Frame generation mode: " << (TimerOn ? "timer" : "idle")
       << "; No particles: " << MySystem->NoParticles();
    DisplayString( 5, glutGet( GLUT_WINDOW_HEIGHT) - 20, os.str(), GLUT_BITMAP_HELVETICA_18);
    
    // Switch back to normal mode (restoring projection)
    glMatrixMode( GL_PROJECTION);
    glPopMatrix();
    glMatrixMode( GL_MODELVIEW);
    glPopMatrix();
    glEnable( GL_TEXTURE_2D);
}


string OpenGLApplication::CommandLabel = "Esc to stop; e-new explosion; t-timer on/ff; h-HUD on/ff";


void OpenGLApplication::DisplayString( const int x, const int y, const string &label, const void * font) const
{
    void * fontToUse = (font != NULL) ? (void*) font : GLUT_BITMAP_HELVETICA_12;
    glRasterPos2i( x, y);
    for ( int i=0; i<(int)label.size(); i++)
        glutBitmapCharacter( fontToUse, label[i]);
}


void OpenGLApplication::MyKeyboard(const unsigned char key, const int x, const int y)
{
    switch (  tolower( key) ) {
        case 27:
            exit(0);
            break;
        case 'e':
            MySystem->Reset();
            break;
        case 'f':
            Fullscreen = !Fullscreen;
            if ( Fullscreen )
                glutFullScreen();
            else
            {
                glutReshapeWindow( glutGet( GLUT_INIT_WINDOW_WIDTH), glutGet( GLUT_INIT_WINDOW_HEIGHT));
                glutPositionWindow( glutGet( GLUT_INIT_WINDOW_X), glutGet( GLUT_INIT_WINDOW_Y));
            }
            break;
        case 't':             // Timer mode / idle mode
            if ( TimerOn )
                glutIdleFunc( idle);
            else
            {
                glutIdleFunc( NULL);
                glutTimerFunc( timerInterval, timer, 0);
            }
            TimerOn = !TimerOn;
            break;
        case 'h':
            ShowHUD = ! ShowHUD;
            break;
        default:
            break;
    }
}


void OpenGLApplication::MyTimer(const int value)
{
    if ( TimerOn )
    {
        glutPostRedisplay();
        glutTimerFunc( timerInterval, timer, 1);
    }
}

int OpenGLApplication::timerInterval = 25;


void OpenGLApplication::MyIdle()
{
    glutPostRedisplay();
}


// main (of course...)

int main( int argc, char ** argv)
{
    OpenGLApplication::Initialize( argc, argv);
    return 0;
}